import random
from pytest import raises

from datadoc_model.Model import DataDocVariable, LanguageStrings
from datadoc_model.Enums import Datatype
from datadoc import state
from datadoc.Enums import SupportedLanguages
from datadoc.DatasetReader import (
    KNOWN_BOOLEAN_TYPES,
    KNOWN_DATETIME_TYPES,
    KNOWN_FLOAT_TYPES,
    KNOWN_INTEGER_TYPES,
    KNOWN_STRING_TYPES,
    DatasetReader,
)
from .utils import TEST_PARQUET_FILEPATH, TEST_SAS7BDAT_FILEPATH


def test_use_abstract_class_directly():
    with raises(TypeError):
        DatasetReader().get_fields()


def test_get_fields_parquet():
    expected_fields = [
        DataDocVariable(short_name="pers_id", data_type=Datatype.STRING),
        DataDocVariable(short_name="tidspunkt", data_type=Datatype.DATETIME),
        DataDocVariable(short_name="sivilstand", data_type=Datatype.STRING),
        DataDocVariable(short_name="alm_inntekt", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="sykepenger", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="ber_bruttoformue", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="fullf_utdanning", data_type=Datatype.STRING),
        DataDocVariable(short_name="hoveddiagnose", data_type=Datatype.STRING),
    ]

    reader = DatasetReader.for_file(TEST_PARQUET_FILEPATH)
    fields = reader.get_fields()

    assert fields == expected_fields


def test_get_fields_sas7bdat():
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    expected_fields = [
        DataDocVariable(
            short_name="tekst",
            name=LanguageStrings(nb="Tekst"),
            data_type=Datatype.STRING,
        ),
        DataDocVariable(
            short_name="tall", name=LanguageStrings(nb="Tall"), data_type=Datatype.FLOAT
        ),
        DataDocVariable(
            short_name="dato",
            name=LanguageStrings(nb="Dato"),
            data_type=Datatype.DATETIME,
        ),
    ]

    reader = DatasetReader.for_file(TEST_SAS7BDAT_FILEPATH)
    fields = reader.get_fields()

    assert fields == expected_fields


def test_get_fields_unknown_file_type():
    with raises(NotImplementedError):
        DatasetReader.for_file("my_dataset.csv").get_fields()


def test_transform_datatype_unknown_type():
    expected = None
    input_data = "definitely not a known data type"
    actual = DatasetReader.transform_data_type(input_data)
    assert actual == expected


def test_transform_datatype():
    for expected, input_options in [
        (Datatype.INTEGER, KNOWN_INTEGER_TYPES),
        (Datatype.FLOAT, KNOWN_FLOAT_TYPES),
        (Datatype.STRING, KNOWN_STRING_TYPES),
        (Datatype.DATETIME, KNOWN_DATETIME_TYPES),
        (Datatype.BOOLEAN, KNOWN_BOOLEAN_TYPES),
    ]:
        input_data = random.choice(input_options)
        actual = DatasetReader.transform_data_type(input_data)
        assert actual == expected



from xml.dom.minidom import parseString
import requests


class VariableDefinition:
    # TODO: Denne skal gå mot ny VarDef, men bruker foreløpig gamle VarDok!
    def __init__(self, vardef_id):
        self.vardef_id = vardef_id
        self.vardef_uri = (
            "https://www.ssb.no/a/xml/metadata/conceptvariable/vardok/"
            + self.vardef_id
            + "/nb"
        )
        self.vardef_gui_uri = (
            "https://www.ssb.no/a/metadata/conceptvariable/vardok/"
            + self.vardef_id
            + "/nb"
        )
        self.vardef_name = None
        self.vardef_definition = None
        self.vardef_short_name = None
        self.get_variable_definition()

    def get_variable_definition(self):
        # TODO: Denne skal gå mot ny VarDef, men bruker foreløpig gamle VarDok!
        vardok_xml = requests.get(self.vardef_uri)
        variable_document = parseString(vardok_xml.text)
        self.vardef_name = variable_document.getElementsByTagName("Title")[
            0
        ].firstChild.nodeValue
        self.vardef_definition = variable_document.getElementsByTagName("Description")[
            0
        ].firstChild.nodeValue



from enum import Enum


class SupportedLanguages(str, Enum):
    "Reference: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry"
    NORSK_BOKMÅL = "nb"
    NORSK_NYNORSK = "nn"
    ENGLISH = "en"



import datetime
import json
import logging
import os
import pathlib
from typing import Dict, Optional

from datadoc_model.Enums import (
    AdministrativeStatus,
    DatasetState,
)
from datadoc.DatasetReader import DatasetReader
from datadoc_model import Model
import datadoc.frontend.DisplayDataset as DisplayDataset
import datadoc.frontend.DisplayVariables as DisplayVariables
from datadoc.utils import calculate_percentage

logger = logging.getLogger(__name__)

OBLIGATORY_DATASET_METADATA = [
    m.identifier for m in DisplayDataset.DISPLAY_DATASET.values() if m.obligatory
]

OBLIGATORY_VARIABLES_METADATA = [
    m.identifier for m in DisplayVariables.DISPLAY_VARIABLES.values() if m.obligatory
]

# These don't vary at runtime so we calculate them as constants here
NUM_OBLIGATORY_DATASET_FIELDS = len(
    [
        k
        for k in Model.DataDocDataSet().dict().keys()
        if k in OBLIGATORY_DATASET_METADATA
    ]
)
NUM_OBLIGATORY_VARIABLES_FIELDS = len(
    [
        k
        for k in Model.DataDocVariable().dict().keys()
        if k in OBLIGATORY_VARIABLES_METADATA
    ]
)


class DataDocMetadata:
    def __init__(self, dataset):
        self.dataset = dataset
        self.dataset_full_path = pathlib.Path(self.dataset)
        self.dataset_directory = self.dataset_full_path.resolve().parent
        self.dataset_name = self.dataset_full_path.name
        self.dataset_stem = self.dataset_full_path.stem  # filename without file ending
        self.metadata_document_name = str(self.dataset_stem) + "__DOC.json"
        self.metadata_document_full_path = self.dataset_directory.joinpath(
            self.metadata_document_name
        )
        self.dataset_state = self.get_dataset_state(self.dataset_full_path)
        self.dataset_version = self.get_dataset_version()
        try:
            self.current_user = os.environ["JUPYTERHUB_USER"]
        except KeyError:
            self.current_user = "default_user@ssb.no"
            logger.warning(
                f"JUPYTERHUB_USER env variable not set, using {self.current_user} as placeholder"
            )
        self.current_datetime = str(datetime.datetime.now())

        self.meta: "Model.MetadataDocument" = Model.MetadataDocument(
            percentage_complete=0,
            document_version=1,
            dataset=Model.DataDocDataSet(),
            variables=[],
        )

        self.variables_lookup: Dict[str, "Model.DataDocVariable"] = {}

        self.read_metadata_document()

    def get_dataset_state(
        self, dataset_path: pathlib.Path = None
    ) -> Optional[DatasetState]:
        """Use the path to attempt to guess the state of the dataset"""

        if dataset_path is None:
            dataset_path = self.dataset_full_path
        dataset_path_parts = list(dataset_path.parts)
        if "kildedata" in dataset_path_parts:
            return DatasetState.SOURCE_DATA
        elif "inndata" in dataset_path_parts:
            return DatasetState.INPUT_DATA
        elif "klargjorte_data" in dataset_path_parts:
            return DatasetState.PROCESSED_DATA
        else:
            return None

    def get_dataset_version(self) -> Optional[str]:
        """Find version information if exists in filename,
        eg. 'v1' in filename 'person_data_v1.parquet'"""
        splitted_file_name = str(self.dataset_stem).split("_")
        if len(splitted_file_name) >= 2:
            last_filename_element = str(splitted_file_name[-1])
            if (
                len(last_filename_element) >= 2
                and last_filename_element[0:1] == "v"
                and last_filename_element[1:].isdigit()
            ):
                return last_filename_element[1:]
        return None

    def read_metadata_document(self):
        fresh_metadata = {}
        if self.metadata_document_full_path.exists():
            with open(self.metadata_document_full_path, "r", encoding="utf-8") as JSON:
                fresh_metadata = json.load(JSON)
            logger.info(
                f"Opened existing metadata file {self.metadata_document_full_path}"
            )

            variables_list = fresh_metadata.pop("variables", None)

            self.meta.variables = [Model.DataDocVariable(**v) for v in variables_list]
            self.meta.dataset = Model.DataDocDataSet(
                **fresh_metadata.pop("dataset", None)
            )
        else:
            self.generate_new_metadata_document()

        self.variables_lookup = {v.short_name: v for v in self.meta.variables}

    def generate_new_metadata_document(self):
        self.ds_schema = DatasetReader.for_file(self.dataset)

        self.meta.dataset = Model.DataDocDataSet(
            short_name=self.dataset_stem,
            assessment=None,
            dataset_state=self.dataset_state,
            name=None,
            data_source=None,
            population_description=None,
            administrative_status=AdministrativeStatus.DRAFT,
            version=self.dataset_version,
            unit_type=None,
            temporality_type=None,
            description=None,
            spatial_coverage_description=None,
            id=None,
            owner=None,
            data_source_path=str(self.dataset_full_path),
            created_date=self.current_datetime,
            created_by=self.current_user,
            last_updated_date=None,
            last_updated_by=None,
        )

        self.meta.variables = self.ds_schema.get_fields()

    def write_metadata_document(self) -> None:
        """Write all currently known metadata to file"""
        json_str = json.dumps(self.meta.dict(), indent=4, sort_keys=False, default=str)
        self.metadata_document_full_path.write_text(json_str, encoding="utf-8")
        logger.info(f"Saved metadata document {self.metadata_document_full_path}")

    @property
    def percent_complete(self) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator."""

        num_all_fields = NUM_OBLIGATORY_DATASET_FIELDS
        num_set_fields = len(
            [
                k
                for k, v in self.meta.dataset.dict().items()
                if k in OBLIGATORY_DATASET_METADATA and v is not None
            ]
        )

        for variable in self.meta.variables:
            num_all_fields += NUM_OBLIGATORY_VARIABLES_FIELDS
            num_set_fields += len(
                [
                    k
                    for k, v in variable.dict().items()
                    if k in OBLIGATORY_VARIABLES_METADATA and v is not None
                ]
            )

        return calculate_percentage(num_set_fields, num_all_fields)



from datadoc.frontend.DisplayDataset import DISPLAY_DATASET, DatasetIdentifiers
from datadoc.frontend.DisplayVariables import DISPLAY_VARIABLES, VariableIdentifiers
from datadoc_model.Model import (
    DataDocDataSet,
    DataDocVariable,
)


def test_dataset_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in DatasetIdentifiers] == list(DataDocDataSet().dict().keys())
    assert [i for i in DatasetIdentifiers] == list(DISPLAY_DATASET.keys())


def test_variables_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in VariableIdentifiers] == list(
        DataDocVariable().dict().keys()
    )
    assert [i for i in VariableIdentifiers] == list(DISPLAY_VARIABLES.keys())



from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type, List, Callable
from pydantic import BaseModel
from dash import dcc
from dash.development.base_component import Component

from datadoc.Enums import SupportedLanguages
from datadoc import state


INPUT_KWARGS = {
    "debounce": True,
    "style": {"width": "100%"},
    "className": "ssb-input",
}
NUMBER_KWARGS = dict(**INPUT_KWARGS, **{"type": "number"})
DROPDOWN_KWARGS = {
    "style": {"width": "100%"},
    "className": "ssb-dropdown",
}


def kwargs_factory():
    """For initialising the field extra_kwargs. We aren't allowed to
    directly assign a mutable type like a dict to a dataclass field"""
    return INPUT_KWARGS


def get_standard_metadata(metadata: BaseModel, identifier: str) -> Any:
    return metadata.dict()[identifier]


def get_multi_language_metadata(
    metadata: BaseModel, identifier: str
) -> Optional[SupportedLanguages]:
    value = getattr(metadata, identifier)
    if value is None:
        return value
    return getattr(value, state.current_metadata_language)


@dataclass
class DisplayMetadata:
    identifier: str
    display_name: str
    description: str
    obligatory: bool = False
    editable: bool = True
    multiple_language_support: bool = False


@dataclass
class DisplayVariablesMetadata(DisplayMetadata):
    options: Optional[Dict[str, List[Dict[str, str]]]] = None
    presentation: Optional[str] = "input"


@dataclass
class DisplayDatasetMetadata(DisplayMetadata):
    extra_kwargs: Dict[str, Any] = field(default_factory=kwargs_factory)
    component: Type[Component] = dcc.Input
    options: Optional[Dict[str, List[Dict[str, str]]]] = None
    value_getter: Callable[[BaseModel, str], Any] = get_standard_metadata



import argparse
import logging
import os
import re
from typing import Any, Dict, List, Tuple, Type

import dash_bootstrap_components as dbc
from dash import ALL, Dash, Input, Output, ctx, dash_table, dcc, html

from datadoc.Enums import SupportedLanguages
import datadoc.state as state
from datadoc.Callbacks import (
    accept_dataset_metadata_input,
    accept_variable_metadata_input,
    update_dataset_metadata_language,
    update_variable_table_language,
)
from datadoc.DataDocMetadata import DataDocMetadata
from datadoc.frontend.DisplayVariables import DISPLAY_VARIABLES
from datadoc.frontend.DisplayDataset import (
    NON_EDITABLE_DATASET_METADATA,
    OBLIGATORY_EDITABLE_DATASET_METADATA,
    OPTIONAL_DATASET_METADATA,
    DisplayDatasetMetadata,
)
from datadoc.utils import running_in_notebook, get_display_values

logger = logging.getLogger(__name__)


DATASET_METADATA_INPUT = "dataset-metadata-input"
COLORS = {"dark_1": "#F0F8F9", "green_1": "#ECFEED", "green_4": "#00824D"}


def build_app(dash_class: Type[Dash], dataset_path: str) -> Dash:

    state.metadata = DataDocMetadata(dataset_path)

    app = dash_class(
        name="DataDoc",
        external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP],
        assets_folder=f"{os.path.dirname(__file__)}/assets",
    )

    def make_dataset_metadata_accordion_item(
        title: str,
        metadata_inputs: List[DisplayDatasetMetadata],
    ) -> dbc.AccordionItem:
        return dbc.AccordionItem(
            title=title,
            children=[
                dbc.Row(
                    [
                        dbc.Col(html.Label(i.display_name)),
                        dbc.Col(
                            i.component(
                                placeholder=i.description,
                                disabled=not i.editable,
                                id={
                                    "type": DATASET_METADATA_INPUT,
                                    "id": i.identifier,
                                },
                                **i.extra_kwargs,
                                **(i.options or {}),
                            ),
                            width=5,
                        ),
                        dbc.Col(width=4),
                    ]
                )
                for i in metadata_inputs
            ],
        )

    def make_ssb_styled_tab(label: str, content: dbc.Container) -> dbc.Tab:
        return dbc.Tab(
            label=label,
            # Replace all whitespace with dashes
            tab_id=re.sub(r"\s+", "-", label.lower()),
            label_class_name="ssb-tabs navigation-item",
            label_style={"margin-left": "10px", "margin-right": "10px"},
            style={"padding": "4px"},
            children=content,
        )

    def make_ssb_warning_alert(
        alert_identifier: str, title: str, content_identifier: str
    ) -> dbc.Alert:
        return dbc.Alert(
            id=alert_identifier,
            is_open=False,
            dismissable=True,
            fade=True,
            class_name="ssb-dialog warning",
            children=[
                dbc.Row(
                    [
                        dbc.Col(
                            width=1,
                            children=[
                                html.Div(
                                    className="ssb-dialog warning icon-panel",
                                    children=[
                                        html.I(
                                            className="bi bi-exclamation-triangle",
                                        ),
                                    ],
                                )
                            ],
                        ),
                        dbc.Col(
                            [
                                html.H5(
                                    title,
                                ),
                                dcc.Markdown(
                                    id=content_identifier,
                                ),
                            ]
                        ),
                    ],
                )
            ],
            color="danger",
        )

    dataset_details = make_ssb_styled_tab(
        "Datasett",
        dbc.Container(
            [
                dbc.Row(html.H2("Datasett detaljer", className="ssb-title")),
                dbc.Accordion(
                    always_open=True,
                    children=[
                        make_dataset_metadata_accordion_item(
                            "Obligatorisk",
                            OBLIGATORY_EDITABLE_DATASET_METADATA,
                        ),
                        make_dataset_metadata_accordion_item(
                            "Valgfritt",
                            OPTIONAL_DATASET_METADATA,
                        ),
                        make_dataset_metadata_accordion_item(
                            "Maskingenerert",
                            NON_EDITABLE_DATASET_METADATA,
                        ),
                    ],
                ),
            ],
        ),
    )

    variables_table = make_ssb_styled_tab(
        "Variabler",
        dbc.Container(
            children=[
                dbc.Row(html.H2("Variabel detaljer", className="ssb-title")),
                dbc.Row(
                    dash_table.DataTable(
                        id="variables-table",
                        # Populate fields with known values
                        data=[
                            get_display_values(v, state.current_metadata_language)
                            for v in state.metadata.meta.variables
                        ],
                        # Define columns based on the information in DISPLAY_VARIABLES
                        columns=[
                            {
                                "name": variable.display_name,
                                "id": variable.identifier,
                                "editable": variable.editable,
                                "presentation": variable.presentation,
                                "hideable": variable.editable,
                            }
                            for variable in DISPLAY_VARIABLES.values()
                        ],
                        # Non-obligatory variables are hidden by default
                        hidden_columns=[
                            v.identifier
                            for v in DISPLAY_VARIABLES.values()
                            if v.obligatory is False
                        ],
                        # Include tooltips for all columns
                        tooltip_header={
                            v.identifier: v.description
                            for v in DISPLAY_VARIABLES.values()
                        },
                        editable=True,
                        # Enable sorting and pagination
                        sort_action="native",
                        page_action="native",
                        # Populate the options for all dropdown values
                        dropdown={
                            v.identifier: v.options
                            for v in DISPLAY_VARIABLES.values()
                            if v.options
                        },
                    )
                ),
            ],
        ),
    )

    header = dbc.CardBody(
        dbc.Row(
            children=[
                # html.Link(rel="stylesheet", href="assets/bundle.css"),
                html.H1("DataDoc", className="ssb-title", style={"color": "white"}),
            ],
        ),
        style={"backgroundColor": COLORS["green_4"]},
    )

    progress_bar = dbc.CardBody(
        style={"padding": "4px"},
        children=[dbc.Progress(id="progress-bar", color=COLORS["green_4"], value=40)],
    )

    controls_bar = dbc.CardBody(
        style={"padding": "4px"},
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Button(
                            [
                                html.I(
                                    className="bi bi-save",
                                    style={"padding-right": "10px"},
                                ),
                                "   Lagre",
                            ],
                            class_name="ssb-btn primary-btn",
                            id="save-button",
                        ),
                    ),
                    dbc.Col(
                        dcc.Dropdown(
                            id="language-dropdown",
                            searchable=False,
                            value=state.current_metadata_language.value,
                            className="ssb-dropdown",
                            options=[
                                {"label": i.name, "value": i.value}
                                for i in SupportedLanguages
                            ],
                        ),
                        align="end",
                        width="auto",
                    ),
                ]
            )
        ],
    )
    dataset_validation_error = make_ssb_warning_alert(
        "dataset-validation-error",
        "Failed validation",
        "dataset-validation-explanation",
    )

    variables_validation_error = make_ssb_warning_alert(
        "variables-validation-error",
        "Failed validation",
        "variables-validation-explanation",
    )

    success_toast = dbc.Alert(
        id="success-message",
        is_open=False,
        dismissable=True,
        fade=True,
        class_name="ssb-dialog",
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        width=3,
                        children=[
                            html.Div(
                                className="ssb-dialog icon-panel",
                                children=[
                                    html.I(
                                        className="bi bi-check-circle",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dbc.Col(
                        [
                            html.H5(
                                "Lagret metadata",
                            ),
                        ]
                    ),
                ],
                align="center",
            )
        ],
        color="success",
    )

    app.layout = dbc.Container(
        style={"padding": "4px"},
        children=[
            header,
            progress_bar,
            controls_bar,
            dbc.CardBody(
                style={"padding": "4px"},
                children=[
                    dbc.Tabs(
                        id="tabs",
                        class_name="ssb-tabs",
                        children=[
                            dataset_details,
                            variables_table,
                        ],
                    ),
                ],
            ),
            variables_validation_error,
            dataset_validation_error,
            success_toast,
        ],
    )

    @app.callback(
        Output("progress-bar", "value"),
        Output("progress-bar", "label"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        Input("variables-table", "data"),
    )
    def callback_update_progress(value, data) -> Tuple[int, str]:
        completion = state.metadata.percent_complete
        return completion, f"{completion}%"

    @app.callback(
        Output("success-message", "is_open"),
        Input("save-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(n_clicks):
        if n_clicks and n_clicks > 0:
            # Write the final completion percentage to the model
            state.metadata.meta.percentage_complete = state.metadata.percent_complete
            state.metadata.write_metadata_document()
            return True
        else:
            return False

    @app.callback(
        Output(
            {"type": DATASET_METADATA_INPUT, "id": ALL},
            "value",
        ),
        Input("language-dropdown", "value"),
    )
    def callback_change_language(language: str):
        return update_dataset_metadata_language(SupportedLanguages(language))

    @app.callback(
        Output("dataset-validation-error", "is_open"),
        Output("dataset-validation-explanation", "children"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(value: Any) -> Tuple[bool, str]:
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"], ctx.triggered_id["id"]
        )

    @app.callback(
        Output("variables-table", "data"),
        Output("variables-validation-error", "is_open"),
        Output("variables-validation-explanation", "children"),
        Input("variables-table", "data"),
        Input("variables-table", "data_previous"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_variable_table(
        data: List[Dict], data_previous: List[Dict], language: str
    ) -> Tuple[List[Dict], bool, str]:
        if ctx.triggered_id == "language-dropdown":
            return update_variable_table_language(data, SupportedLanguages(language))
        else:
            return accept_variable_metadata_input(data, data_previous)

    return app


def main(dataset_path: str = None):
    if dataset_path is None:
        parser = argparse.ArgumentParser()
        parser.add_argument("--dataset-path", help="Specify the path to a dataset")
        args = parser.parse_args()
        # Use example dataset if nothing specified
        dataset = args.dataset_path or "./klargjorte_data/person_data_v1.parquet"
    else:
        dataset = dataset_path

    if running_in_notebook():
        logging.basicConfig(level=logging.WARNING)
        from jupyter_dash import JupyterDash

        JupyterDash.infer_jupyter_proxy_config()
        app = build_app(JupyterDash, dataset)
        app.run_server(mode="inline")
    else:
        logging.basicConfig(level=logging.DEBUG)
        # Assume running in server mode is better (largely for development purposes)
        logger.debug("Starting in development mode")
        app = build_app(Dash, dataset)
        app.run_server(debug=True)


if __name__ == "__main__":
    main()



import pathlib
from abc import ABC, abstractmethod
import pyarrow.parquet as pq
from typing import List, Optional, TypeVar
import pandas as pd
from datadoc import state

from datadoc_model.Enums import Datatype
from datadoc_model.Model import DataDocVariable, LanguageStrings

TDatasetReader = TypeVar("TDatasetReader", bound="DatasetReader")

KNOWN_INTEGER_TYPES = (
    "int",
    "int_",
    "int8",
    "int16",
    "int32",
    "int64",
    "integer",
    "long",
    "uint",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
)

KNOWN_FLOAT_TYPES = (
    "double",
    "float",
    "float_",
    "float16",
    "float32",
    "float64",
    "decimal",
    "number",
    "numeric",
    "num",
)

KNOWN_STRING_TYPES = (
    "string",
    "str",
    "char",
    "varchar",
    "varchar2",
    "text",
    "txt",
    "bytes",
)

KNOWN_DATETIME_TYPES = (
    "timestamp",
    "timestamp[us]",
    "timestamp[ns]",
    "datetime64",
    " datetime64[ns]",
    " datetime64[us]",
    "date",
    "datetime",
    "time",
)

KNOWN_BOOLEAN_TYPES = ("bool", "bool_", "boolean")


class DatasetReader(ABC):
    def __init__(self, dataset):
        self.dataset = dataset

    @staticmethod
    def for_file(dataset: str) -> TDatasetReader:
        """Factory method to return the correct subclass based on the given dataset file"""
        file_type = str(pathlib.Path(dataset)).lower().split(".")[1]
        if file_type == "parquet":
            return DatasetReaderParquet(dataset)
        if file_type == "sas7bdat":
            return DatasetReaderSas7bdat(dataset)
        else:
            # In the future we can potentially support csv, xml or json files
            raise NotImplementedError

    @staticmethod
    def transform_data_type(data_type: str) -> Optional[Datatype]:
        v_data_type = data_type.lower()
        if v_data_type in KNOWN_INTEGER_TYPES:
            return Datatype.INTEGER
        elif v_data_type in KNOWN_FLOAT_TYPES:
            return Datatype.FLOAT
        elif v_data_type in KNOWN_STRING_TYPES:
            return Datatype.STRING
        elif v_data_type in KNOWN_DATETIME_TYPES:
            return Datatype.DATETIME
        elif v_data_type in KNOWN_BOOLEAN_TYPES:
            return Datatype.BOOLEAN
        else:
            return None  # Unknown datatype?

    @abstractmethod
    def get_fields(self) -> List[DataDocVariable]:
        """Abstract method, must be implemented by subclasses"""
        pass


class DatasetReaderParquet(DatasetReader):
    def __init__(self, dataset):
        super().__init__(dataset)

    def get_fields(self) -> List[DataDocVariable]:
        fields = []
        data_table = pq.read_table(self.dataset)
        for data_field in data_table.schema:
            fields.append(
                DataDocVariable(
                    short_name=data_field.name,
                    data_type=self.transform_data_type(str(data_field.type)),
                )
            )
        return fields


class DatasetReaderSas7bdat(DatasetReader):
    def __init__(self, dataset):
        super().__init__(dataset)

    def get_fields(self) -> List[DataDocVariable]:
        fields = []
        # Use an iterator to avoid reading in the entire dataset
        sas_reader = pd.read_sas(self.dataset, iterator=True)

        # Get the first row from the iterator
        row = next(sas_reader)

        # Get all the values from the row and loop through them
        for i, v in enumerate(row.values.tolist()[0]):
            fields.append(
                DataDocVariable(
                    short_name=sas_reader.columns[i].name,
                    # Assume labels are defined in the default language (NORSK_BOKMÅL)
                    # If this is not correct, the user may fix it via the UI
                    name={state.current_metadata_language: sas_reader.columns[i].label},
                    # Access the python type for the value and transform it to a DataDoc Data type
                    data_type=self.transform_data_type(type(v).__name__.lower()),
                )
            )

        return fields



from typing import TYPE_CHECKING
from datadoc.Enums import SupportedLanguages

if TYPE_CHECKING:
    # This is only needed for a type hint so we put the import inside
    # this check to avoid circular imports
    from datadoc.DataDocMetadata import DataDocMetadata

# DANGER: This global is safe when Datadoc is run as designed, with
# an individual instance per user run within a Jupyter Notebook.
#
# If Datadoc is redeployed as a multi-user web app then this storage
# strategy must be modified, since users will modify each others data.
# See here: https://dash.plotly.com/sharing-data-between-callbacks

# Global metadata container
metadata: "DataDocMetadata"

current_metadata_language: SupportedLanguages = SupportedLanguages.NORSK_BOKMÅL



from datadoc_model import Model
from datadoc.Enums import SupportedLanguages


def running_in_notebook() -> bool:
    """Return True if running in Jupyter Notebook"""
    try:
        return get_ipython().__class__.__name__ == "ZMQInteractiveShell"  # type: ignore
    except NameError:
        # The get_ipython method is globally available in ipython interpreters
        # as used in Jupyter. However it is not available in other python
        # interpreters and will throw a NameError. Therefore we're not running
        # in Jupyter.
        return False


def calculate_percentage(completed: int, total: int) -> int:
    """Calculate percentage as a rounded integer"""
    return round((completed / total) * 100)


def get_display_values(
    variable: Model.DataDocVariable, current_language: SupportedLanguages
) -> dict:
    """Return a dictionary representation of Model.DataDocVariable with strings in
    the currently selected language"""
    return_dict = {}
    for field_name, value in variable:
        if isinstance(value, Model.LanguageStrings):
            value = value.dict()[current_language.value]
        return_dict[field_name] = value
    return return_dict



from copy import copy
import os
from pathlib import PurePath
import shutil
import pytest
from datadoc_model import Enums

from datadoc_model.Enums import DatasetState
from datadoc import DataDocMetadata
from datadoc_model.Model import DataDocDataSet, DataDocVariable, MetadataDocument
from .utils import (
    TEST_EXISTING_METADATA_FILE_NAME,
    TEST_EXISTING_METADATA_FILEPATH,
    TEST_PARQUET_FILEPATH,
    TEST_RESOURCES_DIRECTORY,
)


@pytest.fixture
def make_paths():
    split_path = list(PurePath(TEST_PARQUET_FILEPATH).parts)
    initial_data = [
        ("kildedata", DatasetState.SOURCE_DATA),
        ("inndata", DatasetState.INPUT_DATA),
        ("klargjorte_data", DatasetState.PROCESSED_DATA),
        ("", None),
    ]
    test_data = []

    # Construct paths with each of the potential options in them
    for to_insert, state in initial_data:
        new_path = copy(split_path)
        new_path.insert(-2, to_insert)
        new_path = PurePath("").joinpath(*new_path)
        test_data.append((new_path, state))

    return test_data


def test_get_dataset_state(make_paths):
    metadata = DataDocMetadata.DataDocMetadata(TEST_PARQUET_FILEPATH)
    for path, expected_result in make_paths:
        actual_state = metadata.get_dataset_state(path)
        assert actual_state == expected_result


def test_get_dataset_state_no_parameter_supplied():
    metadata = DataDocMetadata.DataDocMetadata(TEST_PARQUET_FILEPATH)
    assert metadata.get_dataset_state() is None


@pytest.fixture
def existing_metadata_file():
    # Setup by copying the file into the relevant directory
    shutil.copy(TEST_EXISTING_METADATA_FILEPATH, TEST_RESOURCES_DIRECTORY)
    yield True  # Dummy value, No need to return anything in particular here
    # Cleanup by deleting the file once we're done
    os.remove(os.path.join(TEST_RESOURCES_DIRECTORY, TEST_EXISTING_METADATA_FILE_NAME))


def test_existing_metadata_file(existing_metadata_file):
    metadata = DataDocMetadata.DataDocMetadata(TEST_PARQUET_FILEPATH)
    assert metadata.meta.dataset.name.en == "successfully_read_existing_file"


def test_metadata_document_percent_complete():
    dataset = DataDocDataSet(dataset_state=Enums.DatasetState.OUTPUT_DATA)
    variable_1 = DataDocVariable(data_type=Enums.Datatype.BOOLEAN)
    variable_2 = DataDocVariable(data_type=Enums.Datatype.INTEGER)
    document = MetadataDocument(
        percentage_complete=0,
        document_version=1,
        dataset=dataset,
        variables=[variable_1, variable_2],
    )
    metadata = DataDocMetadata.DataDocMetadata(TEST_PARQUET_FILEPATH)
    metadata.meta = document

    assert metadata.percent_complete == 11



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datadoc', 'datadoc.frontend', 'datadoc.tests']

package_data = \
{'': ['*'],
 'datadoc': ['assets/*'],
 'datadoc.tests': ['resources/*', 'resources/existing_metadata_file/*']}

install_requires = \
['dash-bootstrap-components>=1.1.0,<2.0.0',
 'dash>=2.4.1,<3.0.0',
 'jupyter-dash>=0.4.2,<0.5.0',
 'pandas>=1.4.2,<2.0.0',
 'pyarrow>=8.0.0,<9.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'ssb-datadoc-model==0.1.0']

entry_points = \
{'console_scripts': ['datadoc = datadoc:main']}

setup_kwargs = {
    'name': 'ssb-datadoc',
    'version': '0.1.1',
    'description': "Document dataset metadata. For use in Statistics Norway's metadata system.",
    'long_description': '# Datadoc\n\nDocument datasets in Statistics Norway\n\n## Usage example\n\n1. Run `from datadoc import main; main("./path/to/your/dataset")` to run Datadoc on a dataset of your choosing.\n1. Complete metadata as you desire\n1. Click `Lagre` to save a metadata document together with your dataset\n\n### If the datadoc package is not installed\n\n1. Clone this repo to your Jupyter instance (or local machine)\n1. Open the `DataDoc.ipynb` Notebook and run the cell to see the example dataset\n\n![DataDoc in use](./doc/change-language-example.gif)\n\n## Contributing\n\n### Dependency Management\n\nPoetry is used for dependency management.\n\nTo install all required dependencies in a virtual environment run `poetry install`. To add a new dependency to the project run `poetry add <package name>`.\n\n### Run project locally in Jupyter\n\nTo run the project locally in Jupyter run:\n\n```bash\npoetry shell\nipython kernel install --user --name="datadoc"\njupyter notebook\n```\n\nA Jupyter instance should open in your browser. Once there, open the `*.ipynb` file. Before running it, select the correct interpreter via `Kernel > Change Kernel > datadoc`.\n\n### Run tests\n\n1. Install dev dependencies (see [Dependency Management](#dependency-management))\n1. Run `poetry shell` to open a shell in the Virtual Environment for the project\n1. Run `pytest` in the root of the project\n',
    'author': 'Statistics Norway',
    'author_email': 'mmw@ssb.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/statisticsnorway/datadoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import Any, Dict, List, Optional, Tuple
from pydantic import ValidationError
from datadoc_model import Model

import datadoc.state as state
from datadoc.Enums import SupportedLanguages
from datadoc.utils import get_display_values
from datadoc.frontend.DisplayDataset import (
    MULTIPLE_LANGUAGE_DATASET_METADATA,
    NON_EDITABLE_DATASET_METADATA,
    OBLIGATORY_EDITABLE_DATASET_METADATA,
    OPTIONAL_DATASET_METADATA,
    DisplayDatasetMetadata,
)
from datadoc.frontend.DisplayVariables import (
    MULTIPLE_LANGUAGE_VARIABLES_METADATA,
    VariableIdentifiers,
)

logger = logging.getLogger(__name__)


def store_language_string(
    metadata_model_object: "Model.DataDocBaseModel",
    value: str,
    metadata_identifier: str,
) -> Optional[Model.LanguageStrings]:
    # In this case we need to set the string to the correct language code
    language_strings = getattr(metadata_model_object, metadata_identifier)
    if language_strings is None:
        # This means that no strings have been saved yet so we need to construct
        # a new LanguageStrings object
        language_strings = Model.LanguageStrings(
            **{state.current_metadata_language.value: value}
        )
    else:
        # In this case there's an existing object so we save this string
        # to the current language
        setattr(language_strings, state.current_metadata_language.value, value)
    return language_strings


def accept_variable_metadata_input(
    data: List[Dict], data_previous: List[Dict]
) -> Tuple[List[Dict], bool, str]:
    updated_row_id = None
    updated_column_id = None
    new_value = None
    show_error = False
    error_explanation = ""
    output_data = data
    update_diff = []
    # What has changed?
    for i in range(len(data)):
        update_diff = list(data[i].items() - data_previous[i].items())
        if update_diff:
            updated_row_id = data[i][VariableIdentifiers.SHORT_NAME.value]
            updated_column_id = update_diff[-1][0]
            new_value = update_diff[-1][-1]
            break  # We're only interested in one change so we break here

    if update_diff:
        try:
            if updated_column_id in MULTIPLE_LANGUAGE_VARIABLES_METADATA:
                if type(new_value) is str:
                    new_value = store_language_string(
                        state.metadata.variables_lookup[updated_row_id],
                        new_value,
                        updated_column_id,
                    )
                elif new_value is None:
                    # This edge case occurs when the user removes the text in an input field
                    # We want to ensure we only remove the content for the current language,
                    # not create a new blank object!
                    new_value = store_language_string(
                        state.metadata.variables_lookup[updated_row_id],
                        "",
                        updated_column_id,
                    )

            logger.debug(
                f"Row: {updated_row_id} Column: {updated_column_id} New value: {new_value}"
            )
            # Write the value to the variables structure
            setattr(
                state.metadata.variables_lookup[updated_row_id],
                updated_column_id,
                new_value,
            )
        except ValidationError as e:
            show_error = True
            error_explanation = f"`{e}`"
            output_data = data_previous
            logger.debug("Caught ValidationError:", exc_info=True)
        else:
            logger.debug(f"Successfully updated {updated_row_id} with {new_value}")

    return output_data, show_error, error_explanation


def accept_dataset_metadata_input(
    value: Any, metadata_identifier: str
) -> Tuple[bool, str]:
    try:
        if (
            metadata_identifier in MULTIPLE_LANGUAGE_DATASET_METADATA
            and type(value) is str
        ):
            value = store_language_string(
                state.metadata.meta.dataset, value, metadata_identifier
            )

        # Update the value in the model
        setattr(
            state.metadata.meta.dataset,
            metadata_identifier,
            value,
        )
    except ValidationError as e:
        show_error = True
        error_explanation = f"`{e}`"
        logger.debug("Caught ValidationError:", exc_info=True)
    else:
        show_error = False
        error_explanation = ""
        logger.debug(f"Successfully updated {metadata_identifier} with {value}")

    return show_error, error_explanation


def update_dataset_metadata_language(language: SupportedLanguages) -> List[Any]:
    """Update the global language setting with the chosen language.
    Return new values for ALL the dataset metadata inputs to allow
    editing of strings in the chosen language"""

    state.current_metadata_language = language
    # The order of this list MUST match the order of display components, as defined
    # in dataset_variables in DataDocDash.py
    displayed_dataset_metadata: List[DisplayDatasetMetadata] = (
        OBLIGATORY_EDITABLE_DATASET_METADATA
        + OPTIONAL_DATASET_METADATA
        + NON_EDITABLE_DATASET_METADATA
    )
    logger.debug(f"Updated language: {state.current_metadata_language.name}")
    return [
        m.value_getter(state.metadata.meta.dataset, m.identifier)
        for m in displayed_dataset_metadata
    ]


def update_variable_table_language(
    data: List[Dict],
    language: SupportedLanguages,
) -> Tuple[List[Dict], bool, str]:
    state.current_metadata_language = language
    new_data = []
    for row in data:
        new_data.append(
            get_display_values(
                state.metadata.variables_lookup[
                    row[VariableIdentifiers.SHORT_NAME.value]
                ],
                state.current_metadata_language,
            )
        )
    logger.debug(f"Updated variable table language: {language.name}")
    return new_data, False, ""



from copy import deepcopy

from datadoc_model.Enums import DatasetState, VariableRole
from datadoc_model.Model import LanguageStrings
from datadoc.tests.utils import TEST_PARQUET_FILEPATH
import datadoc.state as state
from datadoc.DataDocMetadata import DataDocMetadata
from datadoc import Callbacks
from datadoc.Enums import SupportedLanguages


DATA_ORIGINAL = [
    {
        "short_name": "pers_id",
        "variable_role": None,
    },
]
DATA_VALID = [
    {
        "short_name": "pers_id",
        "variable_role": "IDENTIFIER",
    },
]
DATA_NONETYPE = [
    {"short_name": "pers_id", "variable_role": "IDENTIFIER", "name": None},
]
DATA_INVALID = [
    {
        "short_name": "pers_id",
        "variable_role": 3.1415,
    },
]


def test_accept_variable_metadata_input_no_change_in_data():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    output = Callbacks.accept_variable_metadata_input(DATA_ORIGINAL, DATA_ORIGINAL)
    assert output[0] == DATA_ORIGINAL
    assert output[1] is False
    assert output[2] == ""


def test_accept_variable_metadata_input_new_data():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    output = Callbacks.accept_variable_metadata_input(DATA_VALID, DATA_ORIGINAL)

    assert (
        state.metadata.variables_lookup["pers_id"].variable_role
        == VariableRole.IDENTIFIER
    )
    assert output[0] == DATA_VALID
    assert output[1] is False
    assert output[2] == ""


def test_accept_variable_metadata_input_incorrect_data_type():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    previous_metadata = deepcopy(state.metadata.meta.variables)
    output = Callbacks.accept_variable_metadata_input(DATA_INVALID, DATA_ORIGINAL)

    assert output[0] == DATA_ORIGINAL
    assert output[1] is True
    assert "validation error for DataDocVariable" in output[2]
    assert state.metadata.meta.variables == previous_metadata


def test_accept_dataset_metadata_input_new_data():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    output = Callbacks.accept_dataset_metadata_input(
        DatasetState.INPUT_DATA, "dataset_state"
    )
    assert output[0] is False
    assert output[1] == ""
    assert state.metadata.meta.dataset.dataset_state == DatasetState.INPUT_DATA


def test_accept_dataset_metadata_input_incorrect_data_type():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    output = Callbacks.accept_dataset_metadata_input(3.1415, "dataset_state")
    assert output[0] is True
    assert "validation error for DataDocDataSet" in output[1]


def test_change_language():
    ENGLISH_NAME = "English Name"
    BOKMÅL_NAME = "Bokmål Name"
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    state.metadata.meta.dataset.name = LanguageStrings(en=ENGLISH_NAME, nb=BOKMÅL_NAME)
    output = Callbacks.update_dataset_metadata_language(SupportedLanguages.NORSK_BOKMÅL)
    assert ENGLISH_NAME not in output
    assert BOKMÅL_NAME in output
    output = Callbacks.update_dataset_metadata_language(SupportedLanguages.ENGLISH)
    assert ENGLISH_NAME in output
    assert BOKMÅL_NAME not in output


def test_nonetype_value_for_language_string():
    ENGLISH_NAME = "English Name"
    BOKMÅL_NAME = "Bokmål Name"
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    state.metadata.variables_lookup["pers_id"].name = LanguageStrings(
        en=ENGLISH_NAME, nb=BOKMÅL_NAME
    )
    state.current_metadata_language = SupportedLanguages.NORSK_NYNORSK
    Callbacks.accept_variable_metadata_input(DATA_NONETYPE, DATA_ORIGINAL)

    assert state.metadata.variables_lookup["pers_id"].name == LanguageStrings(
        en=ENGLISH_NAME, nb=BOKMÅL_NAME
    )



import os


TEST_RESOURCES_DIRECTORY = "datadoc/tests/resources/"
TEST_PARQUET_FILEPATH = os.path.join(TEST_RESOURCES_DIRECTORY, "person_data_v1.parquet")
TEST_SAS7BDAT_FILEPATH = os.path.join(TEST_RESOURCES_DIRECTORY, "sasdata.sas7bdat")
TEST_EXISTING_METADATA_FILE_NAME = "person_data_v1__DOC.json"
TEST_EXISTING_METADATA_FILEPATH = os.path.join(
    TEST_RESOURCES_DIRECTORY, "existing_metadata_file", TEST_EXISTING_METADATA_FILE_NAME
)
